$Env:CONDA_EXE = "/opt/anaconda1anaconda2anaconda3/bin/conda"
$Env:_CONDA_ROOT = "/opt/anaconda1anaconda2anaconda3"
$Env:_CONDA_EXE = "/opt/anaconda1anaconda2anaconda3/bin/conda"

Import-Module "$Env:_CONDA_ROOT/shell/condabin/Conda.psm1"